/**
 * An interface for the FileSystem.
 *
 * Thanks to:
 *  http://sourceforge.net/apps/mediawiki/predef/index.php?title=Operating_Systems
 *  http://sourceforge.net/apps/mediawiki/predef/index.php?title=Standards
 *  http://pubs.opengroup.org/onlinepubs/009695399/basedefs/unistd.h.html
 *  http://www.cplusplus.com/forum/general/11104/#msg52338
 *  http://www.cplusplus.com/forum/general/11104/#msg52356
 *  http://lists.apple.com/archives/darwin-dev/2008/Dec/msg00037.html
 */

#include "filesystem.h"

#include <climits>
#include <sstream>

#include <boost/filesystem/operations.hpp>
namespace fs = boost::filesystem;

/* ** Platform-dependent code ** */
/*
 * We need to use unistd.h.
 * We already use boost to do some checks on the filesystem, so it isn't
 * a problem using it to know if we can use unistd.h.
 */
#include <boost/config/select_platform_config.hpp>
#include <boost/config/posix_features.hpp>

#ifdef _WIN32
	// Windows
#	define WIN32_LEAN_AND_MEAN 1
#	include <Windows.h>
#elif defined(_hpux) || defined(hpux) || defined(__hpux)
	// HP-UX
#	define _PSTAT64
#	include <sys/pstat.h>
#	include <sys/types.h>
#elif defined(__APPLE__)
	// Mac
#	include <mach-o/dyld.h>
#	include <unistd.h>
#	include <libgen.h>
#endif

/* ** FileSystem class ** */
FileSystem* FileSystem::msSingleton;

/**
 * Return the singleton pointer.
 *
 * It should be always a good pointer because no class should destroy it
 * and if the singleton hasn't been instanced yet, the function do it.
 */
FileSystem *FileSystem::getSingletonPtr()
{
	if(!msSingleton) {
		msSingleton = new FileSystem();
	}
	return msSingleton;
}

/**
 * Return a singleton reference.
 *
 * This function should be safe because it uses getSingletonPtr.
 * See also the comment of this function.
 */
FileSystem &FileSystem::getSingleton()
{
	return *getSingletonPtr();
}

/**
 * The constructor. Initializes the variables.
 */
FileSystem::FileSystem() {
	mArgc = 0;
	mArgv = 0;
}

/**
 * Set the program canonical path from main arguments.
 *
 * Getting the executable name from argv[0] is the most portable way, but
 * it isn't a standard.
 */
void FileSystem::fromArguments(int argc, char *argv[])
{
	if(argc < 1) {
		return;
	}

	// Use boost to check if it's the executable path
	fs::path path(argv[0]);
	if(fs::exists(path)) {
		path = fs::canonical(path);
		mFromArgs = path.generic_string();
	}
}

/**
 * Set the program absolute path using an OS-Depentend way.
 */
fs::path FileSystem::osDependent()
{
	fs::path ret;
	fs::path check;

#	ifdef _WIN32
		ret = std::string(GetModuleFileName(NULL));
#	elif defined(_hpux) || defined(hpux) || defined(__hpux)
		char result[ PATH_MAX ];
		pst_status ps;

		/*
		 * With C it's granted that if the first fails, the second isn't
		 * executed, so it's ok doing so.
		 */
		if(pstat_getproc(&ps, sizeof(ps), 0, getpid()) >= 0 &&
				pstat_getpathname(result, PATH_MAX, &ps.pst_fid_text) >= 0) {
			ret = result;
		}
#	elif defined(__APPLE__)
		char pathbuf[PATH_MAX + 1];
		int bufsize = sizeof(pathbuf);
		_NSGetExecutablePath(pathbuf, &bufsize);
		ret = pathbuf;
#	elif defined(__linux__) || defined(linux) || defined(__linux)
		check = "/proc/self/exe";
		if(fs::exists(check)) {
			ret = fs::read_symlink(check);
		}
#	elif defined(_POSIX_VERSION)
		std::ostringstream stream;

		// Actually just one time loop
		while(true) {
			// BSD
			stream << "/proc/" << getpid() << "/file";
			check = stream.str();
			if(fs::exists(check)) {
				ret = fs::read_symlink(check);
				break;
			}

			// Solaris
			stream.str("");
			stream << "/proc/" << getpid() << "/a.out";
			if(fs::exists(check)) {
				ret = fs::read_symlink(check);
				break;
			}

			break;
		}
#	endif

	return ret;
}

/**
 * Set the absolute root path to the app directory using various methods.
 *
 * Please note that we rely most on the path of the executable file, that is
 * take for granted to be in a subdirectory of the app root path.
 */
const fs::path &FileSystem::appRoot()
{
	if(!mPath.empty()) {
		return mPath;
	}

	mPath = osDependent();

	if(mPath.empty() || !fs::exists(mPath)) {
		mPath = mFromArgs;
	}

	if(!mPath.empty() && fs::exists(mPath)) {
		int i;
		for(i = 0; i < LEVELS; i++) {
			mPath = mPath.parent_path();
		}
		mPath = fs::canonical(mPath);
	} else {
		// Our last hope, but very probably wrong
		mPath = fs::current_path();
	}

	return mPath;
}
